#ifndef _WEBDAV_MOUNT_H_INCLUDE
#define _WEBDAV_MOUNT_H_INCLUDE


/* Copyright (c) 1995 NeXT Computer, Inc. All Rights Reserved */
/*
 * Copyright (c) 1992, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * The NEXTSTEP Software License Agreement specifies the terms
 * and conditions for redistribution.
 *
 *	@(#)webdav_mount.h	8.4 (Berkeley) 1/21/94
 */


struct activate_args
{
	int proxy_ok;								/* ok to use proxy */
	int socket;									/* socket for connection */
};

#endif /*ifndef _WEBDAV_MOUNT_H_INCLUDE */
